@@ -86,7 +86,8 @@ attempt_reconnect (MyApp *app)
 	if (gconf_client_get_bool (app->client,
 				CONFBASE"/auto_retry", NULL) &&
 			! phonemgr_listener_connected (app->listener) &&
-			! app->connecting) {
+			! app->connecting &&
+			app->status == PHONEMGR_LISTENER_IDLE) {
 		g_message ("Auto-retrying the connection");
 		reconnect_phone (app);
 	}
@@ -139,7 +140,8 @@ connect_phone (MyApp *app)
 {
 	g_mutex_lock (app->connecting_mutex);
 	if (phonemgr_listener_connected (app->listener) == FALSE
-	    && app->connecting == FALSE) {
+	    && app->connecting == FALSE
+	    && app->status == PHONEMGR_LISTENER_IDLE) {
 		app->connecting = TRUE;
 		g_mutex_unlock (app->connecting_mutex);
 		/* we're neither connected, nor connecting */
@@ -165,6 +167,8 @@ on_status (PhonemgrListener *listener, int status, MyApp *app)
 	g_message ("Status %d", status);
 	gdk_threads_enter ();
 
+	app->status = status;
+
 	switch (status) {
 		case PHONEMGR_LISTENER_IDLE:
 			g_message ("Disconnect complete");
@@ -183,6 +187,7 @@ on_status (PhonemgrListener *listener, int status, MyApp *app)
 		case PHONEMGR_LISTENER_DISCONNECTING:
 			g_message ("Closing serial port connection");
 			phonemgr_object_emit_number_batteries_changed (app->object, 0);
+			phonemgr_object_emit_network_registration_changed (app->object, 0, 0, 0, 0);
 			break;
 		case PHONEMGR_LISTENER_ERROR:
 			set_icon_state (app);
@@ -223,6 +228,12 @@ on_battery (PhonemgrListener *listener, int percent, gboolean on_ac, MyApp *app)
 	phonemgr_object_emit_battery_state_changed (app->object, 0, percent, on_ac);
 }
 
+static void
+on_network (PhonemgrListener *listener, int mcc, int mnc, int lac, int cid, MyApp *app)
+{
+	phonemgr_object_emit_network_registration_changed (app->object, mcc, mnc, lac, cid);
+}
+
 void
 reconnect_phone (MyApp *app)
 {
@@ -271,13 +282,10 @@ initialise_connection (MyApp *app)
 			G_CALLBACK (on_message), (gpointer) app);
 	app->battery_cb = g_signal_connect (G_OBJECT (app->listener), "battery",
 						G_CALLBACK (on_battery), (gpointer) app);
-#if GLIB_CHECK_VERSION(2,13,0)
+	app->network_cb = g_signal_connect (G_OBJECT (app->listener), "network",
+					    G_CALLBACK (on_network), app);
 	app->reconnector = g_timeout_add_seconds (20, (GSourceFunc)attempt_reconnect,
 						  (gpointer)app);
-#else
-	app->reconnector = g_timeout_add (20 * 1000, (GSourceFunc)attempt_reconnect,
-					  (gpointer)app);
-#endif
 
 	gdk_threads_init ();
 	app->connecting_mutex = g_mutex_new ();
